from __future__ import annotations

from typing import Optional, Any

import disruptive
import disruptive.logging as dtlog
import disruptive.requests as dtrequests
import disruptive.outputs as dtoutputs


class DataConnector(dtoutputs.OutputBase):
    """
    Represents a Data Connector.

    When a Data Connector response is received, the content
    is unpacked and the related attributes are set.

    Attributes
    ----------
    data_connector_id : str
        Unique Data Connector ID.
    project_id : str
        Unique ID of the project where the Data Connector resides.
    display_name : str
        The provided display name.
    data_connector_type : str
        Data Connector :ref:`type <data_connector_types>`.
    status : str
        Whether the Data Connector :ref:`status <data_connector_status>` is
        "ACTIVE", "USER_DISABLED", or "SYSTEM_DISABLED".
    config : HttpPushConfig, None
        An object representing its type-specific
        :ref:`configuration <data_connector_configs>`.
        If an unknown Data Connector :ref:`type <data_connector_types>`
        is receiver, this is None.
    event_Types : list[str]
        List of event types that should be forwarded.
        If empty, all event types are forwarded.
    labels : list[str]
        Device labels are not included by default and will only be forwarded
        with the event if the key is specified in this list.

    """

    # Constants for the various Data Connector configuration types.
    HTTP_PUSH = 'HTTP_PUSH'
    DATA_CONNECTOR_TYPES = [HTTP_PUSH]

    def __init__(self, data_connector: dict) -> None:
        """
        Constructs the Data Connector object by unpacking
        the raw Data Connector response.

        Parameters
        ----------
        data_connector : dict
            Unmodified Data Connector response dictionary.

        """

        # Inherit from OutputBase parent.
        dtoutputs.OutputBase.__init__(self, data_connector)

        # Unpack attributes from dictionary.
        self.data_connector_id: str = data_connector['name'].split('/')[-1]
        self.project_id: str = data_connector['name'].split('/')[1]
        self.status: str = data_connector['status']
        self.display_name: str = data_connector['displayName']
        self.event_types: list[str] = data_connector['events']
        self.labels: list[str] = data_connector['labels']
        self.data_connector_type: str = data_connector['type']
        self.config = self._from_dict(data_connector)

    @classmethod
    def get_data_connector(cls,
                           data_connector_id: str,
                           project_id: str,
                           **kwargs: Any,
                           ) -> DataConnector:
        """
        Gets the current state of a single Data Connector.

        Parameters
        ----------
        data_connector_id : str
            Unique Data Connector ID.
        project_id : str
            Unique project ID.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        data_connector : DataConnector
            Object representing the target Data Connector.

        Examples
        --------
        >>> # Fetch information about a specific Data Connector.
        >>> dcon = disruptive.DataConnector.get_data_connector(
        ...     data_connector_id='<DATA_CONNECTOR_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, data_connector_id)

        # Return DataConnector object of GET request response.
        return cls(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def list_data_connectors(cls,
                             project_id: str,
                             **kwargs: Any,
                             ) -> list[DataConnector]:
        """
        Gets a list of the current state of all Data Connectors in a project.

        Parameters
        ----------
        project_id : str
            Unique project ID.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        data_connectors : list[DataConnector]
            List of objects each representing a Data Connector.

        Examples
        --------
        >>> # List information about all Data Connectors in a project.
        >>> dcons = disruptive.DataConnector.list_data_connectors(
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Return list of DataConnector objects of paginated GET response.
        data_connectors = dtrequests.DTRequest.paginated_get(
            url='/projects/{}/dataconnectors'.format(project_id),
            pagination_key='dataConnectors',
            **kwargs,
        )
        return [cls(dcon) for dcon in data_connectors]

    @classmethod
    def create_data_connector(cls,
                              project_id: str,
                              config: disruptive.DataConnector.HttpPushConfig,
                              display_name: str = '',
                              status: str = 'ACTIVE',
                              event_types: list[str] = [],
                              labels: list[str] = [],
                              **kwargs: Any,
                              ) -> DataConnector:
        """
        Creates a new Data Connector in the specified project.

        Parameters
        ----------
        project_id : str
            Unique project ID.
        config : HttpPushConfig
            An object representing the type-specific
            :ref:`configuration <data_connector_configs>`.
        display_name : str, optional
            Sets a display name for the project.
        status : {"ACTIVE", "USER_DISABLED"} str, optional
            :ref:`Status <data_connector_status>` of the new Data Connector.
        event_types : list[str], optional
            List of event types the Data Connector should forward.
        labels : list[str], optional
            List of labels to forward with each event.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        data_connector : DataConnector
            Object representing the newly created Data Connector.

        Examples
        --------
        >>> # Create a new Data Connector.
        >>> dcon = disruptive.DataConnector.create_data_connector(
        ...     project_id='<PROJECT_ID>',
        ...     config=disruptive.DataConnector.HttpPushConfig(
        ...         url='<HTTPS_ENDPOINT_URL>',
        ...         signature_secret='some-good-secret',
        ...     ),
        ...     display_name='my-first-dcon',
        ... )

        """

        # Construct request body dictionary.
        body: dict = dict()
        body['status'] = status
        body['events'] = event_types
        body['labels'] = labels
        if len(display_name) > 0:
            body['displayName'] = display_name

        # Add the appropriate field depending on config.
        body['type'] = config.data_connector_type
        key, value = config._to_dict()
        body[key] = value

        # Construct URL.
        url = '/projects/{}/dataconnectors'.format(project_id)

        # Return DataConnector object of POST request response.
        return cls(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def update_data_connector(
        cls,
        data_connector_id: str,
        project_id: str,
        config: Optional[disruptive.DataConnector.HttpPushConfig] = None,
        display_name: Optional[str] = None,
        status: Optional[str] = None,
        event_types: Optional[list[str]] = None,
        labels: Optional[list[str]] = None,
        **kwargs: Any,
    ) -> DataConnector:
        """
        Updates the attributes of a Data Connector.
        All parameters that are provided will be updated.

        Parameters
        ----------
        data_connector_id : str
            Unique ID of the Data Connector to update.
        project_id : str
            Unique ID of the project that contains the Data Connector.
        config : HttpPushConfig, optional
            An object representing the type-specific
            :ref:`configuration <data_connector_configs>` for
            the Data Connector.
        display_name : str, optional
            Sets a display name for the Data Connector.
        status : {"ACTIVE", "USER_DISABLED"} str, optional
            :ref:`Status <data_connector_status>` of the Data Connector.
        event_types : list[str], optional
            List of :ref:`event types <event_types>` the
            Data Connector should forward.
        labels : list[str], optional
            List of labels to forward with each event.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        data_connector : DataConnector
            Object representing the updated Data Connector.

        Examples
        --------
        >>> # Update only the display_name of a Data Connector.
        >>> dcon = disruptive.DataConnector.update_data_connector(
        ...     data_connector_id='<DATA_CONNECTOR_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     display_name='new-name',
        ... )

        >>> # Update both the display_name, labels, and forwarded event types.
        >>> dcon = disruptive.DataConnector.update_data_connector(
        ...     data_connector_id='<DATA_CONNECTOR_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     display_name='new-name',
        ...     labels=['room-number', 'customer-id'],
        ...     event_types=[
        ...         disruptive.events.TOUCH,
        ...         disruptive.events.TEMPERATURE,
        ...         ],
        ... )

        """

        # Construct request body dictionary.
        body: dict = dict()
        if display_name is not None:
            body['displayName'] = display_name
        if status is not None:
            body['status'] = status
        if event_types is not None:
            body['events'] = event_types
        if labels is not None:
            body['labels'] = labels

        # Add the appropriate field depending on config.
        if config is not None:
            key, value = config._to_dict()
            body[key] = value

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, data_connector_id)

        # Return DataConnector object of PATCH request response.
        return cls(dtrequests.DTRequest.patch(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def delete_data_connector(cls,
                              data_connector_id: str,
                              project_id: str,
                              **kwargs: Any,
                              ) -> None:
        """
        Deletes the specified Data Connector.

        Parameters
        ----------
        data_connector_id : str
            Unique ID of the Data Connector to delete.
        project_id : str
            Unique ID of the project that contains the Data Connector.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Delete the specified Data Connector.
        >>> disruptive.DataConnector.delete_data_connector(
        ...     data_connector_id='<DATA_CONNECTOR_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, data_connector_id)

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @classmethod
    def get_metrics(cls,
                    data_connector_id: str,
                    project_id: str,
                    **kwargs: Any,
                    ) -> Metric:
        """
        Get the metrics of the last 3 hours for a Data Connector.

        Parameters
        ----------
        data_connector_id : str
            Unique ID of the Data Connector to list metrics for.
        project_id : str
            Unique ID of the project that contains the Data Connector.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        metric : Metric
            Object representing the fetched metrics.

        Examples
        --------
        >>> # Get the 3h metrics of the specified Data Connector.
        >>> metrics = disruptive.DataConnector.get_metrics(
        ...     data_connector_id='<DATA_CONNECTOR_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, data_connector_id)
        url += ':metrics'

        # Return Metric object of GET request response.
        return Metric(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def sync_data_connector(cls,
                            data_connector_id: str,
                            project_id: str,
                            **kwargs: Any,
                            ) -> None:
        """
        Synchronizes the current Data Connector state.

        This method let's you synchronize your cloud service with the current
        state of the devices in your project. This entails pushing the most
        recent event of each type for all devices in your project.

        Parameters
        ----------
        data_connector_id : str
            Unique ID of the Data Connector to synchronize.
        project_id : str
            Unique ID of the project that contains the Data Connector.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Synchronize the specified Data Connector.
        >>> disruptive.DataConnector.sync_data_connector(
        ...     data_connector_id='<DATA_CONNECTOR_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/dataconnectors/{}'
        url = url.format(project_id, data_connector_id)
        url += ':sync'

        # Send POST request, but return nothing.
        dtrequests.DTRequest.post(
            url=url,
            **kwargs,
        )

    @classmethod
    def _from_dict(cls, data_connector: dict) -> Optional[HttpPushConfig]:
        # Isolate the Data Connector type.
        data_connector_type = data_connector['type']

        # Select the appropriate config depending on type.
        if data_connector_type == 'HTTP_PUSH':
            # Isolate config field.
            config = data_connector['httpConfig']

            # Create and return an HttpPush object.
            return cls.HttpPushConfig(
                url=config['url'],
                signature_secret=config['signatureSecret'],
                headers=config['headers'],
            )
        else:
            # If this else statement runs, no config is available for type.
            dtlog.warning('No config available for {} Data Connectors.'.format(
                data_connector_type
            ))
            return None

    class HttpPushConfig():
        """
        Type-specific configurations for the HTTP_PUSH Data Connector.

        Attributes
        ----------
        url : str
            Endpoint URL towards which events are forwarded. Must be HTTPS.
        signature_secret : str
            Secret with which each forwarded event is signed.
        headers : dict[str, str]
            Dictionary of headers to include with each forwarded event.

        """

        data_connector_type = 'HTTP_PUSH'

        def __init__(self,
                     url: Optional[str] = None,
                     signature_secret: Optional[str] = None,
                     headers: Optional[dict] = None,
                     ) -> None:
            """
            Constructs the HttpPushConfig object.

            Parameters
            ----------
            url : str, optional
                Endpoint URL towards which events are forwarded. Must be HTTPS.
            signature_secret : str, optional
                Secret with which each forwarded event is signed.
            headers : dict[str, str], optional
                Dictionary of headers to include with each forwarded event.

            """

            # Set parameter attributes.
            self.url = url
            self.signature_secret = signature_secret
            self.headers = headers

        def _to_dict(self) -> tuple[str, dict]:
            config: dict = dict()
            if self.url is not None:
                config['url'] = self.url
            if self.signature_secret is not None:
                config['signatureSecret'] = self.signature_secret
            if self.headers is not None:
                config['headers'] = self.headers
            return 'httpConfig', config


class Metric(dtoutputs.OutputBase):
    """
    Represents the metrics for a Data Connector over the last 3 hours.

    Attributes
    ----------
    success_count : int
        Number of 2xx responses.
    error_count : int
        Number of non-2xx responses.
    latency : str
        Average latency.

    """

    def __init__(self, metric: dict) -> None:
        """
        Constructs the Metric object by unpacking the raw response.

        """

        # Inherit attributes from ResponseBase parent.
        dtoutputs.OutputBase.__init__(self, metric)

        # Unpack attributes from dictionary.
        self.success_count = metric['metrics']['successCount']
        self.error_count = metric['metrics']['errorCount']
        self.latency = metric['metrics']['latency99p']



from __future__ import annotations

from datetime import datetime
from typing import Optional, Any

import disruptive.requests as dtrequests
import disruptive.outputs as dtoutputs
import disruptive.transforms as dttrans


class ServiceAccount(dtoutputs.OutputBase):
    """
    Represents a Service Account.

    When a Service Account response is received, the content is
    unpacked and the related attributes are updated.

    Attributes
    ----------
    service_account_id : str
        Unique Service Account ID.
    email : str
        Unique Service Account email.
    display_name : str
        The provided display name.
    basic_auth_enabled : bool
        True if Basic Auth is enabled, otherwise False.
    create_time : datetime
        Timestamp of when the Service Account was created.
    update_time : datetime
        Timestamp of when the Service Account was last updated.

    """

    def __init__(self, service_account: dict) -> None:
        """
        Constructs the ServiceAccount object by unpacking the raw response.

        Parameters
        ----------
        service_account : dict
            Unmodified Service Account response dictionary.

        """

        # Inherit from Response parent.
        dtoutputs.OutputBase.__init__(self, service_account)

        # Unpack attributes from dictionary.
        self.service_account_id: str = service_account['name'].split('/')[-1]
        self.email: str = service_account['email']
        self.display_name: str = service_account['displayName']
        self.basic_auth_enabled: bool = service_account['enableBasicAuth']
        self.create_time: Optional[datetime] = \
            dttrans.to_datetime(service_account['createTime'])
        self.update_time: Optional[datetime] = \
            dttrans.to_datetime(service_account['updateTime'])

    @classmethod
    def get_service_account(cls,
                            service_account_id: str,
                            project_id: str,
                            **kwargs: Any,
                            ) -> ServiceAccount:
        """
        Gets the current state of a single Service Account.

        Parameters
        ----------
        service_account_id : str
            Unique ID of the target Service Account.
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        service_account : ServiceAccount
            Object representing the target Service Account.

        Examples
        --------
        >>> # Fetch information about a specific Service Account.
        >>> sa = disruptive.ServiceAccount.get_service_account(
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}'.format(
            project_id,
            service_account_id,
        )

        # Return ServiceAccount object of GET request response.
        return cls(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def list_service_accounts(cls,
                              project_id: str,
                              **kwargs: Any,
                              ) -> list[ServiceAccount]:
        """
        Gets a list of the current state of all Service Accounts in a project.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        service_accounts : list[ServiceAccount]
            List of objects each representing a Service Account.

        Examples
        --------
        >>> # Fetch a list of all Service Accounts in a project.
        >>> sas = disruptive.ServiceAccount.list_service_accounts(
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts'.format(project_id)

        # Return list of ServiceAccount objects of paginated GET response.
        service_accounts = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='serviceAccounts',
            **kwargs,
        )
        return [cls(sa) for sa in service_accounts]

    @classmethod
    def create_service_account(cls,
                               project_id: str,
                               display_name: str = '',
                               basic_auth_enabled: bool = False,
                               **kwargs: Any,
                               ) -> ServiceAccount:
        """
        Create a new Service Account in the specified project.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        display_name : str, optional
            Sets a display name for the Service Account.
        basic_auth_enabled : bool, optional
            Enables Basic Auth for the Service Account if True.
            Defaults to False.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        service_account : ServiceAccount
            Object representing the newly created Service Account.

        Examples
        --------
        >>> # Create a new Service Account with basic auth enabled.
        >>> sa = disruptive.ServiceAccount.create_service_account(
        ...     project_id='<PROJECT_ID>',
        ...     display_name='new-service-account',
        ...     basic_auth_enabled=True,
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts'.format(project_id)

        # Construct body.
        body: dict = dict()
        body['enableBasicAuth'] = basic_auth_enabled
        if len(display_name) > 0:
            body['displayName'] = display_name

        # Return ServiceAccount object of GET request response.
        return cls(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def update_service_account(cls,
                               service_account_id: str,
                               project_id: str,
                               display_name: Optional[str] = None,
                               basic_auth_enabled: Optional[bool] = None,
                               **kwargs: Any,
                               ) -> ServiceAccount:
        """
        Updates the attributes of a specified Service Account.

        Parameters
        ----------
        service_account_id : str
            Unique ID of the target Service Account.
        project_id : str
            Unique ID of the target project.
        display_name : str, optional
            Updates the Service Account display name.
        basic_auth_enabled : bool, optional
            If True, enables Basic Auth while False disables it.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        service_account : ServiceAccount
            Object representing the updated Service Account.

        Examples
        --------
        >>> # Update only the `display_name` of a Service Account.
        >>> sa = disruptive.ServiceAccount.update_service_account(
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     display_name='new-name',
        ... )

        >>> # Update both `display_name` and `basic_auth_enabled`.
        >>> sa = disruptive.ServiceAccount.update_service_account(
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     display_name='new-name',
        ...     basic_auth_enabled=False,
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}'.format(
            project_id,
            service_account_id,
        )

        # Construct body.
        body: dict = dict()
        if display_name is not None:
            body['displayName'] = display_name
        if basic_auth_enabled is not None:
            body['enableBasicAuth'] = basic_auth_enabled

        # Return ServiceAccount object of GET request response.
        return cls(dtrequests.DTRequest.patch(
            url=url,
            body=body,
            **kwargs,
        ))

    @classmethod
    def delete_service_account(cls,
                               service_account_id: str,
                               project_id: str,
                               **kwargs: Any,
                               ) -> None:
        """
        Deletes the specified Service Account.

        Parameters
        ----------
        service_account_id : str
            Unique ID of the Service Account to delete.
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Delete a single Service Account.
        >>> disruptive.ServiceAccount.delete_service_account(
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}'.format(
            project_id,
            service_account_id,
        )

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @staticmethod
    def get_key(key_id: str,
                service_account_id: str,
                project_id: str,
                **kwargs: Any,
                ) -> Key:
        """
        Get the key of a Service Account.

        Parameters
        ----------
        key_id : str
            Unique ID of the target key.
        service_account_id : str
            Unique ID of the target Service Account.
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        key : Key
            Object representing the target key.

        Examples
        --------
        >>> # Get information about a specific key.
        >>> key = disruptive.ServiceAccount.get_key(
        ...     key_id='<KEY_ID>',
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys/{}'.format(
            project_id,
            service_account_id,
            key_id,
        )

        # Return Key object of GET request response.
        return Key(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @staticmethod
    def list_keys(service_account_id: str,
                  project_id: str,
                  **kwargs: Any,
                  ) -> list[Key]:
        """
        Get a list of all keys for a Service Account.

        Parameters
        ----------
        service_account_id : str
            Unique ID of the target Service Account.
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        keys : list[Key]
            List of objects each representing a key.

        Examples
        --------
        >>> # List all keys for a specific Service Account.
        >>> keys = disruptive.ServiceAccount.list_keys(
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys'.format(
            project_id,
            service_account_id,
        )

        # Return list of Key objects of paginated GET response.
        keys = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='keys',
            **kwargs,
        )
        return [Key(key) for key in keys]

    @staticmethod
    def create_key(service_account_id: str,
                   project_id: str,
                   **kwargs: Any,
                   ) -> Key:
        """
        Create a new key for the specified Service Account.

        Parameters
        ----------
        service_account_id : str
            Unique ID of the target Service Account.
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        key : Key
            Object representing the newly created key.

        Examples
        --------
        >>> # Create a new key for a specific Service Account.
        >>> key = disruptive.ServiceAccount.create_key(
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys'.format(
            project_id,
            service_account_id,
        )

        # Return Key object of POST request response.
        response = dtrequests.DTRequest.post(
            url=url,
            **kwargs,
        )
        return Key._with_secret(response)

    @staticmethod
    def delete_key(key_id: str,
                   service_account_id: str,
                   project_id: str,
                   **kwargs: Any,
                   ) -> None:
        """
        Deletes a key in the specified Service Account.

        Parameters
        ----------
        key_id : str
            Unique ID of the key to delete.
        service_account_id : str
            Unique ID of the target Service Account.
        project_id : str
            Unique ID of the target project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Delete a specific key on a Service Account.
        >>> disruptive.ServiceAccount.delete_key(
        ...     key_id='<KEY_ID',
        ...     service_account_id='<SERVICE_ACCOUNT_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}/serviceaccounts/{}/keys/{}'.format(
            project_id,
            service_account_id,
            key_id,
        )

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )


class Key(dtoutputs.OutputBase):
    """
    Represents a key in a Service Account.

    When a key response is received, the content is
    unpacked and the related attributes are updated.

    Attributes
    ----------
    key_id : str
        Unique key ID.
    secret : str, None
        If the Key object was constructed from a newly created key, i.e. from
        calling the :ref:`create_key() <create_key>` method, this attributes
        contains the key secret. This is displayed once, and is otherwise None.
    create_time : datetime
        Timestamp of when the key was created.

    """

    def __init__(self, key: dict) -> None:
        """
        Constructs the Key object by unpacking the raw key response.

        Parameters
        ----------
        key : dict
            Key response dictionary.

        """
        # Inherit from Response parent.
        dtoutputs.OutputBase.__init__(self, key)

        # Initialize secret, which is only not-None when created.
        self.secret: Optional[str] = None

        # Unpack attributes from dictionary.
        self.key_id: str = key['id']
        self.create_time: Optional[datetime] = \
            dttrans.to_datetime(key['createTime'])
        if 'secret' in key:
            self.secret = key['secret']

    @classmethod
    def _with_secret(cls, key: dict) -> Key:
        """
        Moves the secret field in the response dictionary.

        This is done for convenience and consistency. For some reason, when
        the secret is included, the raw response dictionary is changed, adding
        another layer outside the key field. This just flattens the structure.

        Parameters
        ----------
        key : dict
            Unmodified key response dictionary.

        Returns
        -------
        flattened : dict
            Modified key response dictionary.

        """

        flattened = key['key']
        flattened['secret'] = key['secret']
        return cls(flattened)



# Metadata
__version__ = "0.6.1"

# Authentication scheme.
from disruptive.authentication import Auth  # noqa

# When auth is an instance of Auth set with the default constructor,
# the package is unauthenticated, throwing an Unauthorized error at call.
# To authorize, either overwrite the global variable `auth` or provide
# each individual method with the same auth object.
#
# Available authentication methods:
# auth = disruptive.Auth.service_account(key_id, secret, email)
#
# Initialize package with unauthenticated object.
default_auth = Auth.unauthenticated()

# Resources.
from disruptive.resources.device import Device  # noqa
from disruptive.resources.data_connector import DataConnector  # noqa
from disruptive.resources.organization import Organization  # noqa
from disruptive.resources.project import Project  # noqa
from disruptive.resources.stream import Stream  # noqa
from disruptive.resources.eventhistory import EventHistory  # noqa
from disruptive.resources.service_account import ServiceAccount  # noqa
from disruptive.resources.role import Role  # noqa
from disruptive.resources.emulator import Emulator  # noqa

# Additional helper modules.
import disruptive.events as events  # noqa

# If set, logs of chosen level and higher are printed to console.
# Available levels are: debug, info, warning, error, critical.
log_level = None

# REST API base URLs of which all endpoints are an expansion.
base_url = 'https://api.disruptive-technologies.com/v2'
emulator_base_url = 'https://emulator.disruptive-technologies.com/v2'

# If a request response contains an error for which a series of retries is
# worth considering, these variable determine how long to wait without an
# answer, and how many times the package should retry before raising an error.
request_timeout = 3  # seconds
request_attempts = 5  # attempts



from __future__ import annotations

import time
import urllib.parse
from typing import Any

import jwt

from disruptive import requests as dtrequests, errors as dterrors


class _AuthRoutineBase(object):

    def __init__(self) -> None:
        # Set default attributes.
        self._expiration: int = 0
        self._token: str = ''

    def _has_expired(self) -> bool:
        """
        Evaluates whether the access token has expired.

        Returns
        -------
        has_expired : bool
            True if the access token has expired, otherwise False.

        """

        if time.time() > self._expiration:
            return True
        else:
            return False

    def get_token(self) -> str:
        """
        Returns the access token.
        If the token has expired, renew it.

        Returns
        -------
        token : str
            Access token added to the request header.

        """

        # Check expiration time.
        if self._has_expired():
            # Renew access token.
            self.refresh()

        return self._token

    def refresh(self) -> None:
        """
        This function does nothing and is overwritten in all
        child classes. It only exists for consistency purposes
        as it is called in get_token().

        """

        pass


class Unauthenticated(_AuthRoutineBase):

    def __init__(self) -> None:
        # Inherit parent class methods and attributes.
        super().__init__()

    def refresh(self) -> None:
        """
        If called, this function does not but raise an error as no
        authentication routine has been called to update the configuration
        variable, nor has an authentication object been provided.

        Raises
        ------
        Unauthorized
            If neither default_auth has been set nor the
            auth kwarg has been provided.

        """
        raise dterrors.Unauthorized(
            'No authentication method set.'
            ' See developer.d21s.com/api/libraries/python/'
            'authentication.html'
        )


class ServiceAccountAuth(_AuthRoutineBase):
    """
    Ensures that the access token is available and up-to-date.

    Attributes
    ----------
    token_endpoint : str
        URL to which the jwt is exchanged for an access token.

    """

    def __init__(self, key_id: str, secret: str, email: str):
        # Inherit parent class methods and attributes.
        super().__init__()

        # Set parameter attributes.
        self._key_id: str = key_id
        self._secret: str = secret
        self._email: str = email

        # Set default URLs.
        self.token_endpoint = 'https://identity.'\
            'disruptive-technologies.com/oauth2/token'

    @property
    def key_id(self) -> str:
        return self._key_id

    @property
    def secret(self) -> str:
        return self._secret

    @property
    def email(self) -> str:
        return self._email

    def __repr__(self) -> str:
        return '{}.{}({}, {}, {})'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            repr(self.key_id),
            repr(self.secret),
            repr(self.email),
        )

    def refresh(self) -> None:
        """
        Refreshes the access token.

        This first exchanges the JWT for an access token, then updates
        the expiration and token attributes with the response.

        """

        response: dict = self._get_access_token()
        self._expiration = time.time() + response['expires_in']
        self._token = 'Bearer {}'.format(response['access_token'])

    def _get_access_token(self) -> dict:
        """
        Constructs and exchanges the JWT for an access token.

        Returns
        -------
        response : dict
            Dictionary containing expiration and the token itself.

        Raises
        ------
        BadRequest
            If the provided credentials could not be used for authentication.

        """

        # Construct the JWT header.
        jwt_headers: dict[str, str] = {
            'alg': 'HS256',
            'kid': self.key_id,
        }

        # Construct the JWT payload.
        jwt_payload: dict[str, Any] = {
            'iat': int(time.time()),         # current unixtime
            'exp': int(time.time()) + 3600,  # expiration unixtime
            'aud': self.token_endpoint,
            'iss': self.email,
        }

        # Sign and encode JWT with the secret.
        encoded_jwt: str = jwt.encode(
            payload=jwt_payload,
            key=self.secret,
            algorithm='HS256',
            headers=jwt_headers,
        )

        # Prepare HTTP POST request data.
        # Note: The requests package applies Form URL-Encoding by default.
        request_data: str = urllib.parse.urlencode({
            'assertion': encoded_jwt,
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer'
        })

        # Exchange the JWT for an access token.
        try:
            access_token_response: dict = dtrequests.DTRequest.post(
                url='',
                base_url=self.token_endpoint,
                data=request_data,
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                skip_auth=True,
            )
        except dterrors.BadRequest:
            # Re-raise exception with more specific information.
            raise dterrors.BadRequest(
                'Could not authenticate with the provided credentials.\n'
                'Read more: https://developer.d21s.com/docs/authentication'
                '/oauth2#common-errors'
            )

        # Return the access token in the request.
        return access_token_response


class Auth():
    """
    Authenticates the API using a factory design pattern.
    The Auth class itself is only for namespacing purposes.

    Depending on the classmethod called, an instance of some
    authentication routine, like ServiceAccountAuth, is returned.

    If no classmethod has been called, and the API is configured with
    an instance of the Auth class itself, exceptions will be raised
    with indicators on how to properly authenticate.

    """

    @staticmethod
    def unauthenticated() -> Unauthenticated:
        return Unauthenticated()

    @classmethod
    def service_account(cls,
                        key_id: str,
                        secret: str,
                        email: str,
                        ) -> ServiceAccountAuth:
        """
        This method uses an OAuth2 authentication flow. With the provided
        credentials, a `JWT <https://jwt.io/>`_ is created and exchanged for
        an access token.

        Parameters
        ----------
        key_id : str
            Unique Service Account key ID.
        secret : str
            Service Account secret.
        email : str
            Unique Service Account email address.

        Returns
        -------
        auth : ServiceAccountAuth
            Object to initialize and maintain authentication to the REST API.

        """

        # Check that credentials are populated strings.
        cls._verify_str_credentials({
            'key_id': key_id,
            'secret': secret,
            'email': email,
        })

        return ServiceAccountAuth(key_id, secret, email)

    @staticmethod
    def _verify_str_credentials(credentials: dict) -> None:
        """
        Verifies that the provided credentials are strings.

        This check is added as people use environment variables, but
        if for instance os.environ.get() does not find one, it silently
        returns None. It's better to just check for it early.

        Parameters
        ----------
        credentials : dict
            Credentials used to authenticate the REST API.

        """

        for key in credentials:
            # Verify credential is type string.
            if isinstance(credentials[key], str):
                # Raise ConfigurationError if string is empty.
                # This typically happens is credentials are fetched from
                # the environment with a fallback to an empty string.
                if len(credentials[key]) == 0:
                    raise dterrors.ConfigurationError(
                        'Authentication credential <{}> is'
                        ' empty string.'.format(key)
                    )

            # If not, raise TypeError.
            else:
                raise dterrors._raise_builtin(
                    TypeError,
                    'Authentication credential <{}> got type <{}>. '
                    'Expected <str>.'.format(
                        key, type(credentials[key]).__name__
                    )
                )



from disruptive.events.events import Event  # noqa
from disruptive.events.events import Touch  # noqa
from disruptive.events.events import Temperature  # noqa
from disruptive.events.events import TemperatureSample  # noqa
from disruptive.events.events import ObjectPresent  # noqa
from disruptive.events.events import Humidity  # noqa
from disruptive.events.events import ObjectPresentCount  # noqa
from disruptive.events.events import TouchCount  # noqa
from disruptive.events.events import WaterPresent  # noqa
from disruptive.events.events import NetworkStatus  # noqa
from disruptive.events.events import NetworkStatusCloudConnector  # noqa
from disruptive.events.events import BatteryStatus  # noqa
from disruptive.events.events import LabelsChanged  # noqa
from disruptive.events.events import ConnectionStatus  # noqa
from disruptive.events.events import EthernetStatus  # noqa
from disruptive.events.events import CellularStatus  # noqa

# Module constants.
from disruptive.events.events import TOUCH  # noqa
from disruptive.events.events import TEMPERATURE  # noqa
from disruptive.events.events import OBJECT_PRESENT  # noqa
from disruptive.events.events import HUMIDITY  # noqa
from disruptive.events.events import OBJECT_PRESENT_COUNT  # noqa
from disruptive.events.events import TOUCH_COUNT  # noqa
from disruptive.events.events import WATER_PRESENT  # noqa
from disruptive.events.events import NETWORK_STATUS  # noqa
from disruptive.events.events import BATTERY_STATUS  # noqa
from disruptive.events.events import LABELS_CHANGED  # noqa
from disruptive.events.events import CONNECTION_STATUS  # noqa
from disruptive.events.events import ETHERNET_STATUS  # noqa
from disruptive.events.events import CELLULAR_STATUS  # noqa



from __future__ import annotations

from typing import Any

import disruptive.requests as dtrequests
from disruptive.outputs import OutputBase, Member


class Organization(OutputBase):
    """
    Represents an organization.

    When an organization response is received, the content
    is unpacked and the related attributes updated.

    Attributes
    ----------
    organization_id : str
        Unique organization ID.
    display_name : str
        The provided display name.

    """

    def __init__(self, organization: dict) -> None:
        """
        Constructs the Organization object by unpacking the raw response.

        Parameters
        ----------
        organization : dict
            Unmodified organization response dictionary.

        """

        # Inherit from OutputBase parent.
        OutputBase.__init__(self, organization)

        # Unpack attributes from dictionary.
        self.organization_id: str = organization['name'].split('/')[-1]
        self.display_name: str = organization['displayName']

    @classmethod
    def get_organization(cls,
                         organization_id: str,
                         **kwargs: Any,
                         ) -> Organization:
        """
        Gets the current state of a single organization.

        Parameters
        ----------
        organization_id : str
            Unique organization ID.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        organization : Organization
            Object representing the specified organization.

        Examples
        --------
        >>> # Fetch information about a specific organization.
        >>> org = disruptive.Organization.get_organization('<ORGANIZATION_ID>')

        """

        # Construct URL
        url = '/organizations/{}'.format(organization_id)

        # Return Organization object of GET request response.
        return cls(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def list_organizations(cls, **kwargs: Any) -> list[Organization]:
        """
        Gets a list of the current state of all available organizations.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        organizations : list[Organization]
            List of objects each representing an organization.

        Examples
        --------
        >>> # Fetch information about all available organizations.
        >>> orgs = disruptive.Organization.list_organizations()

        """

        # Return list of Organization objects of paginated GET response.
        orgs = dtrequests.DTRequest.paginated_get(
            url='/organizations',
            pagination_key='organizations',
            **kwargs,
        )
        return [cls(org) for org in orgs]

    @staticmethod
    def list_members(organization_id: str,
                     **kwargs: Any,
                     ) -> list[Member]:
        """
        Gets a list of the current state of all members in an organization.

        Parameters
        ----------
        organization_id : str
            Unique ID of the organization to get members from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        members : list[Member]
            List of objects each representing a member.

        Examples
        --------
        >>> # Fetch information about all members in an organization.
        >>> members = disruptive.Organization.list_members('<ORGANIZATION_ID>')

        """

        # Construct URL
        url = '/organizations/{}/members'.format(organization_id)

        # Return list of Member objects of paginated GET response.
        members = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='members',
            **kwargs,
        )
        return [Member(m) for m in members]

    @staticmethod
    def add_member(organization_id: str,
                   email: str,
                   roles: list[str],
                   **kwargs: Any,
                   ) -> Member:
        """
        Add a new member to the specified organization.

        Parameters
        ----------
        organization_id : str
            Unique ID of the organization to add a member to.
        email : str
            Email of the user or Service Account to be added.
        roles : {["organization.admin"]} list[str]
            The role(s) to provide the new member in the organization.
            Currently only supports organization.admin.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        member : Member
            Object representing the newly added member.

        Examples
        --------
        >>> # Add a new member to an organization.
        >>> member = disruptive.Organization.add_member(
        ...     organization_id='<ORGANIZATION_ID>',
        ...     email='<EMAIL_ADDRESS>',
        ...     roles=[disruptive.Role.ORGANIZATION_ADMIN],
        ... )

        """

        # Construct URL
        url = '/organizations/{}/members'.format(organization_id)

        # Construct request body.
        body: dict = dict()
        body['roles'] = ['roles/' + r for r in roles]
        body['email'] = email

        # Return Member object of POST request response.
        return Member(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @staticmethod
    def get_member(member_id: str,
                   organization_id: str,
                   **kwargs: Any,
                   ) -> Member:
        """
        Get a member from the specified organization.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        organization_id : str
            Unique ID of the organization to get a member from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        member : Member
            Object representing the member.

        Examples
        --------
        >>> # Fetch information about a specific member in an organization.
        >>> org = disruptive.Organization.get_organization(
        ...     member_id='<MEMBER_ID>',
        ...     organization_id='<ORGANIZATION_ID>',
        ... )

        """

        # Construct URL
        url = '/organizations/{}/members/{}'.format(
            organization_id,
            member_id,
        )

        # Return Member object of GET request response.
        return Member(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @staticmethod
    def remove_member(member_id: str,
                      organization_id: str,
                      **kwargs: Any,
                      ) -> None:
        """
        Revoke a member's membership in the specified organization.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        organization_id : str
            Unique ID of the organization to remove a member from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Revoke the membership of a member in an organization.
        >>> org = disruptive.Organization.remove_member(
        ...     member_id='<MEMBER_ID>',
        ...     organization_id='<ORGANIZATION_ID>',
        ... )

        """

        # Construct URL
        url = '/organizations/{}/members/{}'.format(
            organization_id,
            member_id,
        )

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @staticmethod
    def get_member_invite_url(member_id: str,
                              organization_id: str,
                              **kwargs: Any,
                              ) -> str:
        """
        Get the invite URL for a member with pending invite.

        This will only work if the invite is still pending. If the invite has
        already been accepted, an error is raised.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        organization_id : str
            Unique ID of the organization of the member to get the URL from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Raises
        ------
        BadRequest
            If the invite has already been accepted.

        Examples
        --------
        >>> # Fetch the pending invite URL for a member.
        >>> org = disruptive.Organization.get_member_invite_url(
        ...     member_id='<MEMBER_ID>',
        ...     organization_id='<ORGANIZATION_ID>',
        ... )

        """

        # Construct URL
        url = '/organizations/{}/members/{}'.format(
            organization_id,
            member_id,
        ) + ':getInviteUrl'

        # Return url string in GET response.
        invite_url: str = dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        )['inviteUrl']
        return invite_url

    @staticmethod
    def list_permissions(organization_id: str,
                         **kwargs: Any,
                         ) -> list[str]:
        """
        List permissions available in the specified organization.

        Parameters
        ----------
        organization_id : str
            Unique ID of the organization to list permissions in.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        permissions : list[str]
            List of available permissions.

        Examples
        --------
        >>> # List all available permissions in an organization.
        >>> org = disruptive.Organization.list_permissions(
        ...     organization_id='<ORGANIZATION_ID>',
        ... )

        """

        # Construct URL
        url = '/organizations/{}/permissions'.format(organization_id)

        # Return list of permissions in GET response.
        permissions: list[str] = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='permissions',
            **kwargs,
        )
        return permissions



from __future__ import annotations

from typing import Optional, Any

import disruptive.requests as dtrequests
from disruptive.outputs import OutputBase, Member


class Project(OutputBase):
    """
    Represents a project.

    When a project response is received, the content is
    unpacked and the related attributes are set.

    Attributes
    ----------
    id : str
        Unique project ID.
    display_name : str
        The provided display name.
    organization_id : str
        Unique ID of parent organization.
    organization_display_name : str
        The provided display name to parent organization.
    sensor_count : int
        Number of sensors in project.
    cloud_connector_count : int
        Number of Cloud Connectors in project.
    is_inventory : bool
        True if project is organization inventory, otherwise False.

    """

    def __init__(self, project: dict) -> None:
        """
        Constructs the Project object by unpacking the raw project response.

        Parameters
        ----------
        project : dict
            Unmodified project response dictionary.

        """

        # Inherit from OutputBase parent.
        OutputBase.__init__(self, project)

        # Unpack attributes from dictionary.
        self.id: str = project['name'].split('/')[-1]
        self.display_name: str = project['displayName']
        self.organization_id: str = project['organization'].split('/')[-1]
        self.organization_display_name: str = \
            project['organizationDisplayName']
        self.sensor_count: int = project['sensorCount']
        self.cloud_connector_count: int = project['cloudConnectorCount']
        self.is_inventory: bool = project['inventory']

    @classmethod
    def get_project(cls,
                    project_id: str,
                    **kwargs: Any,
                    ) -> Project:
        """
        Gets the current state of a single project.

        Parameters
        ----------
        project_id : str
            Unique project ID.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        project : Project
            Object representing the specified project.

        Examples
        --------
        >>> # Fetch information about a specific project.
        >>> project = disruptive.Project.get_project(
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/{}'.format(project_id)

        # Return Project object of GET request response.
        return cls(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @classmethod
    def list_projects(cls,
                      organization_id: Optional[str] = None,
                      query: Optional[str] = None,
                      **kwargs: Any,
                      ) -> list[Project]:
        """
        Gets a list of the current state of all available projects.

        Parameters
        ----------
        organization_id : str
            Unique organization ID.
        query : str, optional
            Keyword based search for project- and organization display names.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        projects : list[Project]
            List of objects each representing a project.

        Examples
        --------
        >>> # Fetch information about all projects in an organization.
        >>> projects = disruptive.Project.list_projects(
        ...     organization_id='<ORGANIZATION_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects'

        # Construct parameters dictionary.
        params = {}
        if organization_id is not None:
            params['organization'] = 'organizations/' + organization_id
        if query is not None:
            params['query'] = query

        # Return list of Project objects of paginated GET response.
        responses = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='projects',
            params=params,
            **kwargs,
        )
        return [cls(r) for r in responses]

    @classmethod
    def create_project(cls,
                       organization_id: str,
                       display_name: str,
                       **kwargs: Any,
                       ) -> Project:
        """
        Create a new project in the specified organization.

        Parameters
        ----------
        organization_id : str
            Unique organization ID.
        display_name : str
            Sets a display name for the project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        project : Project
            Object representing the newly created project.

        Examples
        --------
        >>> # Create a new project.
        >>> project = disruptive.Project.create_project(
        ...     organization_id='<ORGANIZATION_ID>',
        ...     display_name='my-new-project',
        ... )

        """

        # Construct URL.
        url = '/projects'

        # Construct request body.
        body: dict = dict()
        body['organization'] = 'organizations/' + organization_id
        body['displayName'] = display_name

        # Return Project object of POST request response.
        return cls(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @staticmethod
    def update_project(project_id: str,
                       display_name: Optional[str] = None,
                       **kwargs: Any,
                       ) -> None:
        """
        Updates the display name a specified project.

        Parameters
        ----------
        project_id : str
            Unique ID of the project to update.
        display_name : str, optional
            If provided, updates the project display name.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Update the display_name of a project.
        >>> disruptive.Project.update_project(
        ...     project_id='<PROJECT_ID>',
        ...     display_name='new-name',
        ... )

        """

        # Construct URL.
        url = '/projects/' + project_id

        # Construct request body.
        body = {}
        if display_name is not None:
            body['displayName'] = display_name

        # Send PATCH request, but return nothing.
        dtrequests.DTRequest.patch(
            url=url,
            body=body,
            **kwargs,
        )

    @staticmethod
    def delete_project(project_id: str,
                       **kwargs: Any,
                       ) -> None:
        """
        Deletes the specified project.

        Only empty projects can be deleted. If the specified project contains
        any devices or Data Connectors, an error is raised.

        Parameters
        ----------
        project_id : str
            Unique ID of the project to delete.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Raises
        ------
        BadRequest
            If the specified project contains devices or Data Connectors.

        Examples
        --------
        >>> # Delete the specified project.
        >>> disruptive.Project.delete_project(
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL.
        url = '/projects/' + project_id

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @staticmethod
    def list_members(project_id: str,
                     **kwargs: Any,
                     ) -> list[Member]:
        """
        Gets a list of the current state of all members in a project.

        Parameters
        ----------
        project_id : str
            Unique ID of the project to get members from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        members : list[Member]
            List of objects each representing a member.

        Examples
        --------
        >>> # List all members in a project.
        >>> members = disruptive.Project.list_members(
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL
        url = '/projects/{}/members'.format(project_id)

        # Return list of Member objects of paginated GET response.
        members = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='members',
            **kwargs,
        )
        members_list: list[Member] = [Member(m) for m in members]
        return members_list

    @staticmethod
    def add_member(project_id: str,
                   email: str,
                   roles: list[str],
                   **kwargs: Any,
                   ) -> Member:
        """
        Add a new member to the specified project.

        Parameters
        ----------
        project_id : str
            Unique ID of the project to add a member to.
        email : str
            Email of the user or Service Account to be added.
        roles : list[str]
            The :ref:`role(s) <role_types>` to provide the
            new member in the project.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        member : Member
            Object representing the newly added member.

        Examples
        --------
        >>> # Add a new project.developer member to a project.
        >>> member = disruptive.Project.add_member(
        ...     project_id='<PROJECT_ID>',
        ...     email='<MEMBER_EMAIL_ADDRESS>',
        ...     roles=[disruptive.Role.PROJECT_DEVELOPER],
        ... )

        """

        # Construct URL
        url = '/projects/{}/members'.format(project_id)

        # Construct request body.
        body: dict = dict()
        body['roles'] = ['roles/' + r for r in roles]
        body['email'] = email

        # Return Member object of POST request response.
        return Member(dtrequests.DTRequest.post(
            url=url,
            body=body,
            **kwargs,
        ))

    @staticmethod
    def get_member(member_id: str,
                   project_id: str,
                   **kwargs: Any,
                   ) -> Member:
        """
        Get the state of a member in the specified project.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        project_id : str
            Unique ID of the project to get a member from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        member : Member
            Object representing the member.

        Examples
        --------
        >>> # Fetch information about the specified member.
        >>> member = disruptive.Project.get_member(
        ...     member_id='<MEMBER_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL
        url = '/projects/{}/members/{}'.format(
            project_id,
            member_id,
        )

        # Return Member object of GET request response.
        return Member(dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        ))

    @staticmethod
    def update_member(member_id: str,
                      project_id: str,
                      roles: list[str],
                      **kwargs: Any,
                      ) -> Member:
        """
        Update the role(s) of the specified member.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        project_id : str
            Unique ID of the project to update a member in.
        roles : list[str]
            List of new :ref:`roles <role_types>` for the specified member.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        member : Member
            Object representing the updated member.

        Examples
        --------
        >>> # Update the role of a member.
        >>> member = disruptive.Project.update_member(
        ...     member_id='<MEMBER_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     roles=[disruptive.Role.PROJECT_USER],
        ... )

        """

        # Construct URL
        url = '/projects/{}/members/{}'.format(
            project_id,
            member_id,
        )

        # Construct request body.
        body: dict = dict()
        if roles is not None:
            body['roles'] = ['roles/' + r for r in roles]

        # Return updated Member object of PATCH request response.
        return Member(dtrequests.DTRequest.patch(
            url=url,
            body=body,
            **kwargs,
        ))

    @staticmethod
    def remove_member(member_id: str,
                      project_id: str,
                      **kwargs: Any,
                      ) -> None:
        """
        Revoke a member's membership in the specified project.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        project_id : str
            Unique ID of the project to remove a member from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Examples
        --------
        >>> # Remove the specified member from a project.
        >>> disruptive.Project.remove_member(
        ...     member_id='<MEMBER_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL
        url = '/projects/{}/members/{}'.format(
            project_id,
            member_id,
        )

        # Send DELETE request, but return nothing.
        dtrequests.DTRequest.delete(
            url=url,
            **kwargs,
        )

    @staticmethod
    def get_member_invite_url(member_id: str,
                              project_id: str,
                              **kwargs: Any,
                              ) -> str:
        """
        Get the invite URL for a member with pending invite.

        This will only work if the invite is still pending. If the invite has
        already been accepted, an error is raised.

        Parameters
        ----------
        member_id : str
            Unique ID of the member to get.
            For Service Account members, this is the Service Account ID.
            For User members, this is the unique User ID.
        project_id : str
            Unique ID of the project of the member to get the URL from.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        invite_url : str
            The invite url for the specified member.

        Raises
        ------
        BadRequest
            If the invite has already been accepted.

        Examples
        --------
        >>> # Fetch the pending invite URL of a member.
        >>> url = disruptive.Project.get_member_invite_url(
        ...     member_id='<MEMBER_ID>',
        ...     project_id='<PROJECT_ID>',
        ... )

        """

        # Construct URL
        url = '/projects/{}/members/{}'.format(
            project_id,
            member_id,
        ) + ':getInviteUrl'

        # Return url string in GET response.
        invite_url: str = dtrequests.DTRequest.get(
            url=url,
            **kwargs,
        )['inviteUrl']
        return invite_url

    @staticmethod
    def list_permissions(project_id: str,
                         **kwargs: Any,
                         ) -> list[str]:
        """
        List permissions available to the caller in the specified project.

        Parameters
        ----------
        project_id : str
            Unique ID of the project to list permissions in.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        permissions : list[str]
            List of permissions available to the caller.

        Examples
        --------
        >>> # List the available permissions in a project.
        >>> permissions = disruptive.Project.list_permissions('<PROJECT_ID>')

        """

        # Construct URL
        url = '/projects/{}/permissions'.format(project_id)

        # Return list of permissions in GET response.
        permissions: list[str] = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='permissions',
            **kwargs,
        )
        return permissions



from __future__ import annotations

from typing import Generator, Optional, Any

import disruptive.requests as dtrequests
from disruptive.events.events import Event


class Stream():
    """
    Contains staticmethods for streaming events.
    Used for namespacing only and thus does not have a constructor

    """

    @staticmethod
    def event_stream(project_id: str,
                     device_ids: Optional[list[str]] = None,
                     label_filters: Optional[dict] = None,
                     device_types: Optional[list[str]] = None,
                     event_types: Optional[list[str]] = None,
                     **kwargs: Any,
                     ) -> Generator:
        """
        Streams events for one, several, or all devices in a project.

        Implements a basic retry-routine. If connection is lost, the stream
        will attempt to reconnect with an exponential backoff. Potential
        lost events while reconnecting are, however, not acocunted for.

        Parameters
        ----------
        project_id : str
            Unique ID of the target project.
        device_ids : list[str], optional
            Only includes events from the specified device(s).
        label_filters : dict[str, Optional[str]], optional
            Filter devices by their labels. Takes
            the form :code:`{"key": "value"}`, or :code:`{"key": None}` to
            allow any label value.
        device_types : list[str], optional
            Only includes events from devices with specified
            :ref:`type(s) <device_type_constants>`.
        event_types : list[str], optional
            Only includes events of the specified :ref:`type(s) <event_types>`.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        stream : Generator
            A python Generator type that yields each new event in the stream.

        Examples
        --------
        >>> # Stream real-time events from all devices in a project.
        >>> for event in dt.Stream.event_stream('<PROJECT_ID>'):
        ...     print(event)

        >>> # Stream real-time events from one device in a project.
        >>> for event in dt.Stream.event_stream(project_id='<PROJECT_ID>',
        ...                                     device_ids=['<DEVICE_ID>'],
        ...                                     ):
        ...     print(event)

        >>> # Stream real-time touch events from a select list of
        >>> # humidity- and touch sensors, but only those with
        >>> # a 'v1' label in room number 99.
        >>> for e in dt.Stream.event_stream(project_id='<PROJECT_ID>',
        ...                                 device_ids=[
        ...                                     '<DEVICE_ID_1>',
        ...                                     '<DEVICE_ID_2>',
        ...                                     '<DEVICE_ID_3>',
        ...                                 ]
        ...                                 event_types=['touch'],
        ...                                 device_types=['humidity', 'touch'],
        ...                                 label_filters={
        ...                                     'v1': None,
        ...                                     'room-number': '99',
        ...                                 },
        ...                                 ):
        ...     print(e.data)

        """

        # Construct parameters dictionary.
        params: dict = dict()
        if device_ids is not None:
            params['device_ids'] = device_ids
        if device_types is not None:
            params['device_types'] = device_types
        if label_filters is not None:
            params['label_filters'] = []
            for key in label_filters:
                if isinstance(label_filters[key], str):
                    string = key + '=' + label_filters[key]
                    params['label_filters'].append(string)
                else:
                    params['label_filters'].append(key)
        if event_types is not None:
            params['event_types'] = event_types

        # Relay generator output.
        url = '/projects/{}/devices:stream'.format(project_id)
        for event in dtrequests.DTRequest.stream(url, params=params, **kwargs):
            yield Event(event)



from __future__ import annotations

from typing import Optional, Any
from datetime import datetime

import disruptive.requests as dtrequests
import disruptive.transforms as dttrans
from disruptive.events.events import Event


class EventHistory():
    """
    Contains staticmethods for streaming events.
    Used for namespacing only and thus does not have a constructor

    """

    @staticmethod
    def list_events(device_id: str,
                    project_id: str,
                    event_types: Optional[list[str]] = None,
                    start_time: Optional[str | datetime] = None,
                    end_time: Optional[str | datetime] = None,
                    **kwargs: Any,
                    ) -> list[Event]:
        """
        Get the event history for a single device.

        Parameters
        ----------
        device_id : str
            Unique ID of the target device.
        project_id : str
            Unique ID of the target project.
        event_types : list[str], optional
            If provided, only the specified
            :ref:`event types <event_types>` are fetched.
        start_time : str, datetime, optional
            Specifies from when event history is fetched.
            Defaults to 24 hours ago.
        end_time : str, datetime, optional
            Specified until when event history is fetched.
            Defaults to now.
        **kwargs
            Arbitrary keyword arguments.
            See the :ref:`Configuration <configuration>` page.

        Returns
        -------
        events : list[Event]
            A list of all events fetched by the call.

        Examples
        --------
        >>> # Fetch all historic events the last 24 hours.
        >>> events = disruptive.EventHistory.list_events(
        ...     device_id='<DEVICE_ID>',
        ...     project_id='<PROJECT_ID',
        ... )

        >>> # Create a datetime object for 7 days ago.
        >>> from datetime import datetime, timedelta
        >>> seven_days_ago = datetime.now() - timedelta(7)
        >>>
        >>> # Fetch objectPresent event history the last 7 days.
        ... events = dt.EventHistory.list_events(
        ...     device_id='<DEVICE_ID>',
        ...     project_id='<PROJECT_ID>',
        ...     event_types=['objectPresent'],
        ...     start_time=seven_days_ago,
        ... )

        """

        # Construct URL.
        url = '/projects/{}/devices/{}/events'.format(project_id, device_id)

        # Construct parameters dictionary.
        params: dict = dict()
        if event_types is not None:
            params['eventTypes'] = event_types

        # Sanitize timestamps as they must be iso8601 format.
        start_time_iso8601 = dttrans.to_iso8601(start_time)
        if start_time_iso8601 is not None:
            params['startTime'] = start_time_iso8601
        end_time_iso8601 = dttrans.to_iso8601(end_time)
        if end_time_iso8601 is not None:
            params['endTime'] = end_time_iso8601

        # Send paginated GET request.
        res = dtrequests.DTRequest.paginated_get(
            url=url,
            pagination_key='events',
            params=params,
            **kwargs,
        )

        # Return list of Event objects of paginated GET response.
        events: list[Event] = Event.from_mixed_list(res)
        return events



from __future__ import annotations

import logging
from datetime import datetime

import disruptive
import disruptive.errors as dterrors

# Fetch the disruptive logger, but with disabled output.
logger = logging.getLogger('disruptive')
logger.setLevel(99)

levels = {'debug': 1, 'info': 2, 'warning': 3, 'error': 4, 'critical': 5}


def debug(msg: str | dict) -> None:
    if _log_flag_exceeds('debug'):
        _fmt_log(msg, 'DEBUG')
    logger.debug(msg)


def info(msg: str | dict) -> None:
    if _log_flag_exceeds('info'):
        _fmt_log(msg, 'INFO')
    logger.info(msg)


def warning(msg: str | dict) -> None:
    if _log_flag_exceeds('warning'):
        _fmt_log(msg, 'WARNING')
    logger.warning(msg)


def error(msg: str | dict) -> None:
    if _log_flag_exceeds('error'):
        _fmt_log(msg, 'ERROR')
    logger.error(msg)


def critical(msg: str | dict) -> None:
    if _log_flag_exceeds('critical'):
        _fmt_log(msg, 'CRITICAL')
    logger.critical(msg)


def _log_flag_exceeds(level: str | dict) -> bool:
    # If None, never True.
    if disruptive.log_level is None:
        return False

    # Verify set value is valid.
    if disruptive.log_level not in levels:
        raise dterrors.ConfigurationError(
            'Invalid logging level {}. '
            'Must be either None, "debug", "info", '
            '"warning", "error", or "critical".'.format(disruptive.log_level)
        )

    # Check if level is exceeded.
    if levels[level] >= levels[disruptive.log_level]:
        return True
    else:
        return False


def _fmt_log(msg: str | dict, level: str | dict) -> None:
    print('[{}] {:<8} - {}'.format(
        datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3],
        level,
        msg,
    ))



from __future__ import annotations

import re
import base64
from datetime import datetime
from typing import Optional

import disruptive.errors as dterrors


def base64_encode(string: str) -> str:
    string_bytes = string.encode('ascii')
    base64_bytes = base64.b64encode(string_bytes)
    base64_string = base64_bytes.decode('ascii')
    return base64_string


def to_iso8601(ts: Optional[str | datetime]) -> Optional[str]:
    # Verify that we even got a string.
    if isinstance(ts, str):
        # As it is a string, verify iso8601 format.
        if validate_iso8601_format(ts):
            # Nothing to do, return as is.
            return ts
        else:
            # Invalid iso8601 format, raise error.
            msg = 'Timestamp format <{}> is invalid iso8601 format.\n' \
                'Example: 2020-01-01T00:00:00Z'.format(ts)
            raise dterrors.FormatError(msg)

    # If not string, datetime is also fine as it can be converted.
    elif isinstance(ts, datetime):
        # Use built-in functions for converting to string.
        dt = ts.isoformat()

        # Add timezone information if lacking.
        if '+' not in dt:
            dt += 'Z'
        elif '+00:00' in dt:
            dt = dt.replace('+00:00', 'Z')

        return dt

    # If ts is None, return None.
    elif ts is None:
        return None

    # If any other type, raise TypeError.
    else:
        msg = 'Got timestamp of type <{}>, expected ' \
            'iso8601 <str> or <datetime>.'.format(type(ts).__name__)
        raise dterrors._raise_builtin(TypeError, msg)


def to_datetime(ts: Optional[str | datetime]) -> Optional[datetime]:
    # Check if input is already datetime format.
    if isinstance(ts, datetime):
        # Nothing to do, just return it.
        return ts

    # If input is string, we might be able to convert it.
    elif isinstance(ts, str):
        # First, verify if string is valid iso8601 format.
        if validate_iso8601_format(ts):
            # Use built-in functions for converting to datetime.
            return datetime.fromisoformat(ts.replace('Z', '+00:00'))
        else:
            # Invalid iso8601 format, raise error.
            msg = f'Timestamp format [{ts}] is invalid iso8601 format.\n' \
                'Example: 2020-01-01T00:00:00Z'
            raise dterrors.FormatError(msg)

    # If ts is None, return None.
    elif ts is None:
        return None

    # If any other type, raise TypeError.
    else:
        msg = 'Got timestamp of type <{}>, expected ' \
            'iso8601 <str> or <datetime>.'.format(
                type(ts).__name__
            )
        raise dterrors._raise_builtin(TypeError, msg)


def validate_iso8601_format(dt_str: str) -> bool:
    # Set up regex for matching iso8601 string.
    # This should probably be changed in the future as it is
    # a little forced. However, the reason for using this approach is
    # that the datetime built-in method for checking iso8601 format
    # allows missing timezone infromation (i.e. Z or +-00:00 suffix).
    # This must be included in our API, and is why this regex exists.
    iso8601_regex = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-' \
        '(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):' \
        '([0-5][0-9])?(.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])$'
    match_iso8601 = re.compile(iso8601_regex).match

    if match_iso8601(dt_str) is not None:
        return True
    else:
        return False


def _celsius_to_fahrenheit(celsius: float) -> float:
    """
    Converts Celsius temperature value to Fahrenheit.

    Parameters
    ----------
    celsius : float
        Temperature value in Celsius.

    Returns
    -------
    fahrenheit : float
        Temperature value in Fahrenheit if Celsius is not None.

    """

    return (celsius * (9/5)) + 32
